"""
cdis_oauth2client.oauth2
"""

import flask
import requests

from .exceptions import OAuth2Error


def get_username(user_api=None):
    """
    Given the URL for the user API, call user-api to get the username for the
    current flask session using the current access_token cookie.

    For this function to get the username, the user must already be
    authenticated with an access_token cookie stored in the flask session:

        flask.session['access_token']

    If the `user_api` argument is not provided, get the user API URL from:

        flask.current_app.config['USER_API']

    :param user_api: URL for the user API
    :type user_api: str
    :return: the username:
    :rtype: str
    """
    if user_api is None:
        try:
            user_api = flask.current_app.config['USER_API']
        except KeyError as e:
            raise OAuth2Error("'USER_API' not set in flask.current_app.config")
    auth_header = flask.request.headers.get('Authorization')
    if auth_header and auth_header.startswith('bearer'):
        access_token = auth_header.split(' ')[1]
    else:
        try:
            access_token = flask.session['access_token']
        except KeyError:
            code = flask.request.args.get('code')
            if code is None:
                raise OAuth2Error('could not obtain access token')
            access_token = flask.current_app.oauth2.get_access_token(code)
    url = user_api + 'user/'
    headers = {'Authorization': 'Bearer ' + access_token}
    try:
        response = requests.get(url, headers=headers).json()
    except requests.RequestException as e:
        msg = 'failed to get username due to requests exception: {}'
        raise OAuth2Error(msg.format(e))
    username = response.get('username')
    if username is None:
        msg = 'username missing from response: {}'
        raise OAuth2Error(msg.format(response))
    return username



"""
cdis_oauth2client.client

Defines the OAuth2Client class for use by CDIS internal microservices.
"""

import urllib.request, urllib.parse, urllib.error

from cdispyutils.log import get_logger
import requests

from .exceptions import OAuth2Error


_PDC = 'https://bionimbus-pdc.opensciencedatacloud.org/api/oauth2/'


class OAuth2Client(object):
    """
    OAuth2 client that can be used by CDIS internal microservices.
    """

    logger = get_logger('OAuth2Client')

    def __init__(
            self, client_id, client_secret, redirect_uri, oauth_provider=_PDC,
            internal_oauth_provider=None, scope='user'):
        """
        According to the Flask-OAuthlib docs, the Client should have (at least)
        the following properties:

        `client_id`
            A random string.

        `client_secret`
            A random string.

        `client_type`
            A string representing if 'it is confidential' (quote from the docs)

        `redirect_uris`
            A list of redirect uris.

        `default_redirect_uri`
            One of the redirect uris to default to.

        `default_scopes`
            The default scopes of the client.

        An example using SQLAlchemy can be found here:
        https://flask-oauthlib.readthedocs.io/en/latest/oauth2.html
        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.oauth_provider = oauth_provider
        self.internal_oauth = internal_oauth_provider or oauth_provider
        self.redirect_uri = redirect_uri
        self.scope = scope

    @property
    def authorization_url(self):
        """Return the URL at which the authorization can be obtained."""
        tail = urllib.parse.urlencode({
            'client_id': self.client_id,
            'redirect_uri': self.redirect_uri,
            'response_type': 'code',
            'scope': self.scope
        })
        return urllib.parse.urljoin(self.oauth_provider, 'authorize') + '?' + tail

    def _post_to_internal_oauth(self, data):
        """
        Post data to the URL at ``self.internal_oauth``.

        This is used for retrieving or refreshing the OAuth credentials (see
        ``get_token`` and ``refresh_token``, which also have examples for what
        ``data`` should look like).
        """
        try:
            url = urllib.parse.urljoin(self.internal_oauth, 'token')
            return requests.post(url, data=data).json()
        except requests.RequestException as e:
            msg = 'request failed to reach oauth provider: ' + str(e.message)
            self.logger.exception(msg)
            return {'error': msg}
        except Exception as e:  # pylint: disable=broad-except
            msg = (
                'unexpected error trying to reach oauth provider: '
                + str(e.message)
            )
            self.logger.exception(msg)
            return {'error': msg}

    def get_token(self, code):
        """
        Retrieve the oauth credentials.

        An example of the credentials returned:

            {
                u'access_token': u'9ydWQi1SqGU82hAGf8M0JoNJbXhxQ1',
                u'expires_in': 3600, u'refresh_token':
                u'Ll6PfksjrCSJHtkEQV41mRRbR4tUxU', u'scope': u'user',
                u'token_type': u'Bearer'
            }

        :param code: usually flask.request.args.get('code')
        :type code: str
        :return: dictionary of OAuth credentials (see above)
        :rtype: dict
        """
        # Formulate the data for posting to the OAuth provider.
        data = {
            'code': code,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'grant_type': 'authorization_code',
            'redirect_uri': self.redirect_uri
        }
        return self._post_to_internal_oauth(data)

    def get_access_token(self, code):
        """
        Get specifically the access_token from the OAuth token response.
        """
        token_response = self.get_token(code)
        access_token = token_response.get('access_token')
        if not access_token:
            raise OAuth2Error(
                message='did not receive access token',
                json=token_response
            )
        return access_token

    def refresh_token(self, refresh_token):
        """
        Refresh the OAuth access token.

        The credentials returned should be the same as the example in
        ``get_token``.
        """
        # Formulate the data for posting to the OAuth provider.
        data = {
            'refresh_token': refresh_token,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'grant_type': 'refresh_token',
            'redirect_uri': self.redirect_uri
        }
        return self._post_to_internal_oauth(data)



"""
cdis_oauth2client.exceptions

Provides the OAuth2Error exception that represents some problem with the
authentication.
"""


class OAuth2Error(Exception):
    """
    An exception to represent a problem with OAuth2 authentication.

    See :py:func:``authorize`` in :py:mod:``oauth2`` for usage examples.
    """

    def __init__(self, message='', json=None):
        super(OAuth2Error, self).__init__(message)
        self.message = message
        self.code = 400
        self.json = json



"""
cdis_oauth2client.blueprint

Provides a Flask blueprint for authorization.

Exports:
    ``blueprint``
        A Flask blueprint which provides the following endpoints:
        * ``/authorization_url``
              returns the authorization url for the current app's OAuth2Client
        * ``/authorize``
              authorize a user using the ``authorize`` function provided from
              cdis_oauth2client.oauth2
        * ``/logout``
              log out a user by clearing the flask session
"""

import flask
from flask import current_app


blueprint = flask.Blueprint('oauth', 'oauth_v0')


@blueprint.route('/authorization_url', methods=['GET'])
def get_authorization_url():
    """Just return the authorization URL (from the client)."""
    return current_app.oauth2.authorization_url


@blueprint.route('/authorize', methods=['GET'])
def do_authorize():
    """
    Call ``authorize`` to authorize the user by storing the access_token cookie
    received from the OAuth2Client.
    """
    code = flask.request.args.get('code')
    flask.session['access_token'] = current_app.oauth2.get_access_token(code)
    return ''


@blueprint.route('/logout', methods=['GET'])
def logout_oauth():
    """
    Log out the user by clearing the Flask session (thus removing the
    access_token from the session).
    """
    flask.session.clear()
    return ''



"""
cdis_oauth2client

Library providing an OAuth2 client model for CDIS microservices.
"""

from .blueprint import blueprint
from .client import OAuth2Client
from .exceptions import OAuth2Error
from .oauth2 import get_username



from setuptools import setup
from subprocess import check_output


def get_version():
    # https://github.com/uc-cdis/dictionaryutils/pull/37#discussion_r257898408
    try:
        tag = check_output(
            ["git", "describe", "--tags", "--abbrev=0", "--match=[0-9]*"]
        )
        return tag.decode("utf-8").strip("\n")
    except Exception:
        raise RuntimeError(
            "The version number cannot be extracted from git tag in this source "
            "distribution; please either download the source from PyPI, or check out "
            "from GitHub and make sure that the git CLI is available."
        )


setup(
    name='cdis_oauth2client',
    version='1.0.0',
    description='Flask blueprint and utilities for oauth2 client',
    url='https://github.com/uc-cdis/cdis_oauth2client',
    license='Apache',
    packages=[
        'cdis_oauth2client',
    ],
    install_requires=[
        'Flask',
        'requests~=2.5',
        'cdispyutils>=1.0.3',
    ],
)
